from typing import Callable, Dict


def extract_callee_args_as_kwargs(func: Callable) -> Dict[str, any]:
    """
    Extract all parameters a function was called with during execution time as kwarg dict.
    Call this from inside a function and this returns all parameters the function was given at run time.
    Handy to reduce boilerplate on funcs with many args
    :param func: The function to get kwargs from.  A function should pass itself as parameter when calling this
    :return: value kwargs of the func at the time of calling the extract_args_as_kwargs function
    """
    import inspect

    # Get the names of the functions parameters
    param_names = list(
        map(lambda x: x.name, inspect.signature(func).parameters.values())
    )
    # We go 2 frame up, one for the dict comprehension, and one for extract_args_as_kwargs func itself
    kwargs = {
        n: inspect.getouterframes(inspect.currentframe())[2].frame.f_locals[n]
        for n in param_names
    }
    return kwargs



from johnsnowlabs.abstract_base.lib_resolver import try_import_lib

if try_import_lib("sparknlp_display", True):
    from sparknlp_display import DependencyParserVisualizer
    from sparknlp_display import NerVisualizer
    from sparknlp_display import EntityResolverVisualizer
    from sparknlp_display import RelationExtractionVisualizer
    from sparknlp_display import AssertionVisualizer
else:
    pass



import ast
import importlib
import inspect
import os
import site
import subprocess
from typing import List

logged_imports = []


def inject_modules(imports: List[str], g):
    # Imports a list of modules in the global scope g
    for module_name in imports:
        try:
            if "." not in module_name:
                # plain module import
                g[module_name] = importlib.import_module(module_name)
                continue
            elif "*" in module_name:
                # Not handled but unlikely to break anytime soon
                continue
            elif " as " in module_name:
                # import with alias
                name = module_name.split(" ")[-1]
                module = module_name.split(" ")[0]
                attr = module.split(".")[-1]
                module = ".".join(module.split(".")[:-1])
            else:
                # import without alias
                attr = module_name.split(".")[-1]
                name = attr
                module = ".".join(module_name.split(".")[:-1])
            # import the module and attach to globals
            g[name] = getattr(importlib.import_module(module), attr)
        except Exception as e:  # ImportError
            if module_name not in logged_imports:
                logged_imports.append(module_name)
                print(f"Warning: Could not import module={module_name} error = {e}")


def collect_all_imports(source_code) -> List[str]:
    # collect all import statements from source code string
    tree = ast.parse(source_code)
    imports = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.asname is None:
                    imports.append(alias.name)
                else:
                    imports.append(f"{alias.name} as {alias.asname}")
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                if alias.asname is None:
                    imports.append(f"{node.module}.{alias.name}")
                else:
                    imports.append(f"{node.module}.{alias.name} as {alias.asname}")
    return imports


def reverse_compatibility_import(script_path, g):
    """
    Fallback when import fails, this will read source code and import 1-by-1 and add to globals of the calling module
    :param script_path: tge script path, should be fetched with __file__ from calling module
    :param g: globals() from calling script into which we inject imports
    """
    inject_modules(
        collect_all_imports(ast.parse(open(script_path, encoding="utf8").read())), g
    )


def try_import(lib):
    try:
        importlib.reload(site)
        globals()[lib] = importlib.import_module(lib)
        importlib.import_module(lib)
    except Exception as _:
        # print(f'Failed to import {lib}')
        return False
    return True


def try_import_in_venv(lib, py_path):
    c = f'{py_path} -c "import {lib}"'
    try:
        result = subprocess.check_output(c, shell=True, stderr=subprocess.STDOUT)
        if result == b"":
            return True
            print("all good!")
        else:
            print(f"Looks like {lib} is missing \n{result}")
            return False
    except:
        print(f"Looks like {lib} is missing")
        return False


def is_running_in_emr():
    """We populate EMR with environment variable"""
    if "JSL_EMR" in os.environ.keys():
        return bool(os.environ["JSL_EMR"])
    return False


def is_running_in_databricks_runtime():
    """ Check if the currently running Python Process is running in Databricks runtime or not
    """
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def env_required_license():
    if (
        "JSL_AWS_PLATFORM" in os.environ.keys()
        and str(os.environ["JSL_AWS_PLATFORM"]) == "1"
    ):
        return False
    if (
        "JSL_AZURE_PLATFORM" in os.environ.keys()
        and str(os.environ["JSL_AZURE_PLATFORM"]) == "1"
    ):
        return False
    return True


def set_py4j_logger_to_error_on_databricks():
    # Only call this when in Databricks
    try:
        if "SKIP_py4j_DISABLE" in os.environ and os.environ["SKIP_py4j_DISABLE"] == "1":
            return
        import logging

        from pyspark.sql import SparkSession

        logger = SparkSession.builder.getOrCreate()._jvm.org.apache.log4j
        logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)
    except:
        pass


def get_folder_of_func(func):
    # Get the file path where the function is defined
    func_file = inspect.getfile(func)

    # Get the directory name from the file path
    func_dir = os.path.dirname(func_file)

    return func_dir



import os
import sys
from abc import ABC
from typing import List, Set, Optional

import pkg_resources

from johnsnowlabs import settings
from johnsnowlabs.abstract_base.lib_resolver import Py4JJslLibDependencyResolverABC
from johnsnowlabs.auto_install.jsl_home import get_install_suite_from_jsl_home
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.py_models.lib_version import LibVersion
from johnsnowlabs.utils.enums import (
    ProductName,
    ProductLogo,
    ProductSlogan,
    SparkVersion,
)
from johnsnowlabs.utils.env_utils import try_import
from johnsnowlabs.utils.pip_utils import (
    install_standard_pypi_lib,
    install_licensed_pypi_lib,
    get_pip_lib_version,
)
from johnsnowlabs.utils.venv_utils import VenvWrapper


class AbstractSoftwareProduct(ABC):
    """Only first degree dependencies may be contained in the hard/licensed/optional/ dependency lists
    Higher degree dependencies will be resolved by iterating the dependency graph
    By default the ABC implements a check_installed based on import.

    """

    name: ProductName
    logo: ProductLogo
    slogan: Optional[ProductSlogan] = None
    hard_dependencies: Set["AbstractSoftwareProduct"] = set()
    licensed_dependencies: Set["AbstractSoftwareProduct"] = set()
    optional_dependencies: Set["AbstractSoftwareProduct"] = set()
    py_module_name: Optional[str] = None
    pypi_name: Optional[str] = None
    # Only defined for JSL libs below
    install_deps: Optional[bool] = True
    compatible_spark_versions: List[SparkVersion]
    latest_version: Optional[LibVersion] = None
    jsl_url_resolver: Optional[Py4JJslLibDependencyResolverABC] = None
    licensed: bool = False
    is_py4j = False
    pypi_name_databricks: Optional[str] = None

    @classmethod
    def get_installed_version_via_import(cls):
        return False

    @classmethod
    def check_installed(
        cls,
        python_exec_path: Optional[str] = sys.executable,
        download_folder: str = None,
    ) -> bool:
        if cls.pypi_name and download_folder:
            for whl in os.listdir(download_folder):
                # whl file names re-name '-' to '_'
                if cls.pypi_name.replace("-", "_") in whl:
                    return True

        if cls.py_module_name and not python_exec_path:
            return try_import(cls.py_module_name)
        elif python_exec_path:
            return VenvWrapper.is_lib_in_py_exec(
                python_exec_path, cls.py_module_name, False
            )

        # print(f'Assuming {cls.name} is installed, no checks defined.')
        return True

    @classmethod
    def check_installed_correct_version(
        cls, python_exec_path: str = sys.executable, download_folder: str = None
    ) -> bool:
        # Only supported for current Py Exec Path, return True otherwise
        if python_exec_path != sys.executable:
            return True
        if download_folder:
            return True
        if not cls.pypi_name:
            return False
        if not cls.latest_version:
            return True
        if not cls.check_installed(
            python_exec_path=python_exec_path, download_folder=download_folder
        ):
            return False
        try:
            if (
                pkg_resources.get_distribution(cls.pypi_name).version
                == cls.latest_version.as_str()
            ):
                # print(f'👌 Installed version for {cls.logo + cls.name} is correct, no changes made.')
                return True
            else:
                # print(f'🤓 Installed version for {cls.logo + cls.name} is incorrect, '
                #       f'should be {cls.latest_version.as_str()} but is {pkg_resources.get_distribution(cls.pypi_name).version} '
                #       f'upgrading the package')

                return False
        except Exception as err:
            v = get_pip_lib_version(lib=cls.pypi_name, py_exec=python_exec_path)
            if v:
                return v.as_str() == cls.latest_version.as_str()
            return False

    @classmethod
    def get_installed_version(
        cls,
        python_exec_path: str = sys.executable,
        download_folder: str = None,
        prefer_pip=False,
        fallback_import=False,
    ) -> bool:
        # Only supported for current Py Exec Path, return True otherwise
        if not prefer_pip:
            try:
                return pkg_resources.get_distribution(cls.pypi_name).version
            except:
                pass
        v = get_pip_lib_version(lib=cls.pypi_name, py_exec=python_exec_path)
        if v:
            return v.as_str()
        else:
            return cls.get_installed_version_via_import()

    @classmethod
    def check_dependencies(cls, python_exec_path=None) -> bool:
        # print(f'Assuming {cls.name} dependencies are fine, no checks defined.')
        return True

    @classmethod
    def health_check(cls) -> bool:
        # print(f'Assuming {cls.name} is ok, no checks defined.')
        return True

    @classmethod
    def install(
        cls,
        secrets: Optional[JslSecrets] = None,
        py_path=sys.executable,
        upgrade=True,
        re_install=False,
        version: Optional[str] = None,
        download_folder: Optional[str] = None,
        include_dependencies: bool = True,
    ) -> bool:
        """
        Install the product with default settings.
        Defaults to Pypi file_name install.

        -m pip download <module> -d path
        """
        if not cls.install_deps:
            include_dependencies = False
        if not version and cls.latest_version:
            version = cls.latest_version
        if cls.pypi_name:
            if cls.is_py4j and settings.enforce_versions:
                #  p4j lib should have jars/wheels for it in ~/.johnsnowlabs
                # Try using suite whl before attempting to install from remote location
                # Unless we toggle enforce_versions=False
                suite = get_install_suite_from_jsl_home(browser_login=False,force_browser=False)
                if cls.name == ProductName.hc.value and suite.hc and suite.hc.py_lib:

                    return install_standard_pypi_lib(
                        os.path.join(settings.py_dir, suite.hc.py_lib.file_name),
                        cls.py_module_name,
                        python_path=py_path,
                        upgrade=upgrade,
                        re_install=re_install,
                        # version=version,
                        download_folder=download_folder,
                        include_dependencies=include_dependencies,
                    )
                elif (
                    cls.name == ProductName.ocr.value and suite.ocr and suite.ocr.py_lib
                ):
                    return install_standard_pypi_lib(
                        os.path.join(settings.py_dir, suite.ocr.py_lib.file_name),
                        cls.py_module_name,
                        python_path=py_path,
                        upgrade=upgrade,
                        re_install=re_install,
                        # version=version,
                        download_folder=download_folder,
                        include_dependencies=include_dependencies,
                    )
                elif (
                    cls.name == ProductName.nlp.value and suite.nlp and suite.nlp.py_lib
                ):
                    return install_standard_pypi_lib(
                        os.path.join(settings.py_dir, suite.nlp.py_lib.file_name),
                        cls.py_module_name,
                        python_path=py_path,
                        upgrade=upgrade,
                        re_install=re_install,
                        # version=version,
                        download_folder=download_folder,
                        include_dependencies=include_dependencies,
                    )
            if secrets and cls.licensed:
                # Licensed is versioned via the secrets
                # Fallback install if we could not find locally
                return install_licensed_pypi_lib(
                    secrets=secrets,
                    pypi_name=cls.pypi_name,
                    module_name=cls.py_module_name,
                    product=cls,
                    py_path=py_path,
                    download_folder=download_folder,
                    include_dependencies=include_dependencies,
                )
            else:
                return install_standard_pypi_lib(
                    cls.pypi_name,
                    cls.py_module_name,
                    python_path=py_path,
                    upgrade=upgrade,
                    re_install=re_install,
                    download_folder=download_folder,
                    version=version,
                    include_dependencies=include_dependencies,
                )
        # raise NotImplemented(f'No install defined for {cls.file_name}')
        return True

    @classmethod
    def install_cli(
        cls,
    ) -> bool:
        """
        Install the product configurable interactive from CLI
        """
        if cls.pypi_name:
            return install_standard_pypi_lib(cls.pypi_name)
        raise NotImplemented(f"No install defined for {cls.name}")



# https://pypi.org/project/databricks-api/
import base64
import json
from os.path import dirname
from typing import Union

from databricks_api import DatabricksAPI

from johnsnowlabs import settings
from johnsnowlabs.py_models.install_info import JvmInstallInfo, PyInstallInfo
from johnsnowlabs.utils.file_utils import path_tail


def dbfs_file_exists(db: DatabricksAPI, path: str):
    try:
        dbfs_ls(db, path)
        return True
    except:
        return False


def dbfs_ls(db: DatabricksAPI, dbfs_path: str):
    return db.dbfs.list(dbfs_path)


def dbfs_rm(
    db: DatabricksAPI,
    dbfs_path: str,
    recursive: bool = False,
):
    print("🧹 Cleaning up old JSL home in ", dbfs_path)
    return db.dbfs.delete(dbfs_path, recursive=recursive)


def copy_from_local_to_hdfs(
    db: DatabricksAPI, local_path: str, dbfs_path: str, overwrite: bool = True
):
    print(f"Copying {local_path} to remote cluster path {dbfs_path}")
    db.dbfs.put(
        path=dbfs_path,
        overwrite=overwrite,
        # contents=None,
        src_path=local_path,
    )


def put_file_on_dbfs(
    db: DatabricksAPI, path: str, content: any, overwrite: bool = True
):
    try:
        directory = dirname(path)
        db.dbfs.mkdirs(directory)
    except:
        pass
    if type(content) != "string":
        content = json.dumps(content)
    content = base64.b64encode(content.encode()).decode()
    db.dbfs.put(path=path, overwrite=overwrite, contents=content)


def get_db_path(local_info: Union[JvmInstallInfo, PyInstallInfo, str]):
    """Get a deterministic path for a JvmInstallInfo or PyInstallInfo or local files on dbfs.
    Always use this method to generate output file path for dbfs.
    """
    if isinstance(local_info, JvmInstallInfo):
        return f"{settings.dbfs_java_dir}/{local_info.file_name}"
    elif isinstance(local_info, PyInstallInfo):
        # Gotta add the suffix or databricks will not pickup the correct version
        return f'{settings.dbfs_py_dir}/{local_info.file_name.split(".")[0]}-py2.py3-none-any.whl'
    elif isinstance(local_info, str):
        if ".py" in local_info:
            return f"{settings.db_py_jobs_dir}/{path_tail(local_info)}"
        elif ".jar" in local_info:
            return f"{settings.db_jar_jobs_dir}/{path_tail(local_info)}"
        elif ".ipynb" in local_info:
            return f"{settings.db_py_notebook_dir}/{path_tail(local_info)}"
        else:
            raise Exception(
                f"Invalid Job file, file name must contain either .py .jar or .ipynb"
                f"But got {local_info}"
            )
    else:
        raise Exception(f"Invalid install type = {type(local_info)}")



import os
import sys
from typing import Any, List

from langchain.embeddings.base import Embeddings

# from langchain.pydantic_v1 import BaseModel, Extra
from pydantic import BaseModel, Extra

from johnsnowlabs.frameworks.embedding_retrieval.utils import (
    get_docsplitter_pipe,
    get_doc_splitter_internal_pipe,
)


class JohnSnowLabsLangChainEmbedder(BaseModel, Embeddings):
    """JohnSnowLabs embedding models

    To use, you should have the ``johnsnowlabs`` python package installed.
    Example:        .. code-block:: python
            from langchain.embeddings.johnsnowlabs import JohnSnowLabsEmbeddings
            document = "foo bar"
            embedding = JohnSnowLabsEmbeddings('embed_sentence.bert')
            output = embedding.embed_query(document)
    """

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        raise NotImplementedError("JohnSnowLabsEmbeddings does not support async yet")

    async def aembed_query(self, text: str) -> List[float]:
        raise NotImplementedError("JohnSnowLabsEmbeddings does not support async yet")

    model: Any

    def __init__(
        self, model="embed_sentence.bert", hardware_target="cpu", **kwargs: Any
    ):
        """Initialize the johnsnowlabs model."""
        super().__init__(**kwargs)
        # 1) Check imports
        try:
            from johnsnowlabs import nlp
            from nlu.pipe.pipeline import NLUPipeline
        except ImportError as exc:
            raise ImportError(
                "Could not import johnsnowlabs python package. "
                "Please install it with `pip install johnsnowlabs`."
            ) from exc

            # 2) Start a Spark Session
        try:
            os.environ["PYSPARK_PYTHON"] = sys.executable
            os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable
            nlp.start(hardware_target=hardware_target)
        except Exception as exc:
            raise Exception("Failure starting Spark Session") from exc
            # 3) Load the model
        try:
            if isinstance(model, str):
                self.model = nlp.load(model)
            elif isinstance(model, NLUPipeline):
                self.model = model
            else:
                self.model = nlp.to_nlu_pipe(model)
        except Exception as exc:
            raise Exception("Failure loading model") from exc

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a JohnSnowLabs transformer model.

        Args:            texts: The list of texts to embed.
        Returns:            List of embeddings, one for each text."""

        return self.model.predict_embeds(texts)

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a JohnSnowLabs transformer model.
        Args:            text: The text to embed.
        Returns:            Embeddings for the text."""
        return self.model.predict_embeds(text)[0]


class JohnSnowLabsLangChainDocumentSplitter:
    def __init__(
        self,
        chunk_size: int = 500,
        chunk_overlap: int = 0,
        split_patterns: list = ["\n\n", "\n", " ", ""],
        explode_splits: bool = True,
        keep_separators: bool = True,
        patterns_are_regex: bool = False,
        trim_whitespace: bool = True,
    ):
        """
        :param chunk_size: Size of each chunk of text.
        :param chunk_overlap: Length of the overlap between text chunks , by default `0`.
        :param split_patterns: Patterns to separate the text by in decreasing priority , by default `["\n\n", "\n", " ", ""]`.
        :param explode_splits: Whether to explode split chunks to separate rows , by default `True`.
        :param keep_separators: Whether to keep the separators in the final result , by default `True`.
        :param patterns_are_regex: Whether to interpret the split patterns as regular expressions , by default `False`.
        :param trim_whitespace: Whether to trim whitespaces of extracted chunks , by default `True`.
        """
        self.pipe = get_docsplitter_pipe(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_patterns=split_patterns,
            explode_splits=explode_splits,
            keep_separators=keep_separators,
            patterns_are_regex=patterns_are_regex,
            trim_whitespace=trim_whitespace,
        )

    def split_documents(self, docs):
        return split_lang_docs(self.pipe, docs)


class JohnSnowLabsLangChainMedicalDocumentSplitter(JohnSnowLabsLangChainDocumentSplitter):
    def __init__(
        self,
        # Splitter params
        split_mode: str = "recursive",
        chunk_size: int = 500,
        chunk_overlap: int = 0,
        split_patterns: list = None,    # different value for split modes
        explode_splits: bool = True,
        keep_separators: bool = True,
        patterns_are_regex: bool = False,
        trim_whitespace: bool = True,
        sentence_awareness: bool = False,
        max_length: int = None,         # different value for split modes
        custom_bounds_strategy: str = "prepend",
        case_sensitive: bool = False,
        meta_data_fields: list = [],
        enable_sentence_increment: bool = False,
        # Tokenizer Params
        tokenizer_target_pattern: str = None,
        tokenizer_prefix_pattern: str = None,
        tokenizer_suffix_pattern: str = None,
        tokenizer_infix_patterns: list = None,
        tokenizer_exceptions: list = None,
        tokenizer_exceptions_path: str = None,
        tokenizer_case_sensitive_exceptions: bool = None,
        tokenizer_context_chars: list = None,
        tokenizer_split_pattern: str = None,
        tokenizer_split_chars: list = None,
        tokenizer_min_length: int = None,
        tokenizer_max_length: int = None,
        # Sentence Detector Params
        sentence_model_architecture: str = None,
        sentence_explode_sentences: bool = None,
        sentence_custom_bounds: list = None,
        sentence_use_custom_bounds_only: bool = None,
        sentence_split_length: int = None,
        sentence_min_length: int = None,
        sentence_max_length: int = None,
        sentence_impossible_penultimates: list = None,
    ):
        """
        :param split_mode: The split mode to determine how text should be segmented. Default: 'recursive'.
            It should be one of the following values:
                - "char": Split text based on individual characters.
                - "token": Split text based on tokens.
                - "sentence": Split text based on sentences.
                - "recursive": Split text recursively using a specific algorithm.
                - "regex": Split text based on a regular expression pattern.
        :param chunk_size: Size of each chunk of text. This param is applicable only for "recursive" split_mode.
        :param chunk_overlap: Length of the overlap between text chunks , by default `0`. This param is applicable only for "recursive" split_mode.
        :param split_patterns: Patterns to separate the text by in decreasing priority , by default `["\n\n", "\n", " ", ""]`.
        :param explode_splits: Whether to explode split chunks to separate rows , by default `True`.
        :param keep_separators: Whether to keep the separators in the final result , by default `True`.
        :param patterns_are_regex: Whether to interpret the split patterns as regular expressions , by default `False`.
        :param trim_whitespace: Whether to trim whitespaces of extracted chunks , by default `True`.
        :param sentence_awareness: Whether to split the document by sentence awareness if possible. Default: False.
            If true, it can stop the split process before maxLength. This param is not applicable only for "regex" and "recursive" split_mode.
        :param max_length: The maximum length allowed for spitting. The mode in which the maximum length is specified:
            - "char": Maximum length is measured in characters. Default: 512
            - "token": Maximum length is measured in tokens. Default: 128
            - "sentence": Maximum length is measured in sentences. Default: 8
        :param custom_bounds_strategy: The custom bounds strategy for text splitting using regular expressions.
            This param is applicable only for "regex" split_mode. It should be one of the following values:
                - "none": No custom bounds are applied.
                - "prepend": Custom bounds are prepended to the split documents.
                - "append": Custom bounds are appended to the split documents.
                - Default: "prepend".
        :param case_sensitive: Whether to use case sensitive when matching regex, by default `False`.
        :param meta_data_fields: Metadata fields to add specified data in columns to the metadata of the split documents.
            You should set column names to read columns.
        :param enable_sentence_increment: Whether the sentence index should be incremented in the metadata of the annotator. Default: False.

        Tokenizer parameters
        --------------------
        :param tokenizer_target_pattern: Pattern to grab from text as token candidates, by default ``\\S+``
        :param tokenizer_prefix_pattern: Regex with groups and begins with ``\\A`` to match target prefix, by default ``\\A([^\\s\\w\\$\\.]*)``
        :param tokenizer_suffix_pattern: Regex with groups and ends with ``\\z`` to match target suffix, by default ``([^\\s\\w]?)([^\\s\\w]*)\\z``
        :param tokenizer_infix_patterns: Regex patterns that match tokens within a single target. groups identify different sub-tokens. multiple defaults
        :param tokenizer_exceptions: Words that won't be affected by tokenization rules
        :param tokenizer_exceptions_path: Path to txt file with list of token exceptions
        :param tokenizer_case_sensitive_exceptions: Whether to care for case sensitiveness in exceptions, by default True
        :param tokenizer_context_chars: Character list used to separate from token boundaries, by default ['.', ',', ';', ':', '!', '?', '*', '-', '(', ')', '"', "'"]
        :param tokenizer_split_pattern: Pattern to separate from the inside of tokens. Takes priority over splitChars.
        :param tokenizer_split_chars: Character list used to separate from the inside of tokens
        :param tokenizer_min_length: The minimum allowed length for each token, by default 0
        :param tokenizer_max_length: The maximum allowed length for each token, by default 99999

        SentenceDetectorDLModel parameters
        ----------------------------------
        :param sentence_model_architecture: Model architecture (CNN)
        :param sentence_explode_sentences: Whether to explode each sentence into a different row, for better parallelization. Defaults to false.
        :param sentence_custom_bounds: Characters used to explicitly mark sentence bounds, by default []
        :param sentence_use_custom_bounds_only: Only utilize custom bounds in sentence detection, by default False
        :param sentence_split_length: Length at which sentences will be forcibly split
        :param sentence_min_length: The minimum allowed length for each sentence, by default 0
        :param sentence_max_length: The maximum allowed length for each sentence, by default 99999
        :param sentence_impossible_penultimates: Impossible penultimates - list of strings which a sentence can't end with
        """
        self.pipe = get_doc_splitter_internal_pipe(
            split_mode=split_mode,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_patterns=split_patterns,
            explode_splits=explode_splits,
            keep_separators=keep_separators,
            patterns_are_regex=patterns_are_regex,
            trim_whitespace=trim_whitespace,
            sentence_awareness=sentence_awareness,
            max_length=max_length,
            custom_bounds_strategy=custom_bounds_strategy,
            case_sensitive=case_sensitive,
            meta_data_fields=meta_data_fields,
            enable_sentence_increment=enable_sentence_increment,
            # Tokenizer Params
            tokenizer_target_pattern=tokenizer_target_pattern,
            tokenizer_prefix_pattern=tokenizer_prefix_pattern,
            tokenizer_suffix_pattern=tokenizer_suffix_pattern,
            tokenizer_infix_patterns=tokenizer_infix_patterns,
            tokenizer_exceptions=tokenizer_exceptions,
            tokenizer_exceptions_path=tokenizer_exceptions_path,
            tokenizer_case_sensitive_exceptions=tokenizer_case_sensitive_exceptions,
            tokenizer_context_chars=tokenizer_context_chars,
            tokenizer_split_pattern=tokenizer_split_pattern,
            tokenizer_split_chars=tokenizer_split_chars,
            tokenizer_min_length=tokenizer_min_length,
            tokenizer_max_length=tokenizer_max_length,
            # Sentence Detector Params
            sentence_model_architecture=sentence_model_architecture,
            sentence_explode_sentences=sentence_explode_sentences,
            sentence_custom_bounds=sentence_custom_bounds,
            sentence_use_custom_bounds_only=sentence_use_custom_bounds_only,
            sentence_split_length=sentence_split_length,
            sentence_min_length=sentence_min_length,
            sentence_max_length=sentence_max_length,
            sentence_impossible_penultimates=sentence_impossible_penultimates,
        )


def split_lang_doc(pipe, doc):
    from langchain.schema.document import Document

    return [
        Document(page_content=split, metadata=doc.metadata)
        for split in pipe.annotate(doc.page_content)["splits"]
    ]


def split_lang_docs(pipe, docs):
    return [split for doc in docs for split in split_lang_doc(pipe, doc)]



import shutil
import urllib.request
from typing import Union, Any
from urllib.request import urlopen

import requests

from johnsnowlabs.abstract_base.pydantic_model import WritableBaseModel
from johnsnowlabs.py_models.lib_version import LibVersion
from johnsnowlabs.utils.enums import JvmHardwareTarget, PyInstallTypes, ProductName
from johnsnowlabs.utils.enums import SparkVersion


class UrlDependency(WritableBaseModel):
    """Representation of a URL"""

    url: str
    dependency_type: Union[JvmHardwareTarget, PyInstallTypes]
    spark_version: SparkVersion
    dependency_version: LibVersion
    file_name: str
    product_name: ProductName

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.file_name = self.url.split("/")[-1]

    def update_url(self, new_url):
        self.url = new_url

    def validate(self):
        # Try GET on the URL and see if its valid/reachable
        return requests.head(self.url).status_code == 200

    @staticmethod
    def internet_on():
        try:
            return True if urlopen("https://www.google.com/", timeout=10) else False
        except:
            return False

    def download_url(
        self, save_path, name_print_prefix: str = "", keep_default_file_name=True
    ):
        if not UrlDependency.internet_on():
            print(
                f"Warning! It looks like there is no active internet connection on this machine "
                f"Trying to continue but might run into problems..."
            )

        if not self.validate():
            raise ValueError(f"Trying to download Invalid URL! {self.url}")
        if keep_default_file_name:
            self.file_name = self.url.split("/")[-1]
        save_path = save_path + "/" + self.file_name

        print(f"Downloading {name_print_prefix} {self.file_name}")
        # Download the file from `url` and save it locally under `file_name`:
        with urllib.request.urlopen(self.url) as response, open(
            save_path, "wb"
        ) as out_file:
            shutil.copyfileobj(response, out_file)
